# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the MBAR estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np

from warnings import warn as _warn
from msmtools.util.exceptions import NotConvergedWarning as _NotConvergedWarning

from thermotools import mbar as _mbar
from .callback import CallbackInterrupt

__all__ = [
    'update_therm_weights',
    'estimate_therm_energies',
    'estimate']

cdef extern from "_mbar_direct.h":
    void _mbar_direct_update_therm_weights(
        int *therm_state_counts, double *therm_weights, double *bias_weight_sequences,
        int n_therm_states, int seq_length, double *new_therm_weights)

def update_therm_weights(
    _np.ndarray[int, ndim=1, mode="c"] therm_state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_weights not None,
    bias_weight_sequences, # _np.ndarray[double, ndim=2, mode="c"]
    _np.ndarray[double, ndim=1, mode="c"] new_therm_weights not None):
    r"""
    Calculate the reduced thermodynamic free energies therm_energies
        
    Parameters
    ----------
    therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        state counts in each of the T thermodynamic states
    therm_weights : numpy.ndarray(shape=(T), dtype=numpy.float64)
        probabilities of the T thermodynamic states
    bias_weight_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        bias weights in the T thermodynamic states for all X samples
    new_therm_weights : numpy.ndarray(shape=(T), dtype=numpy.float64)
        target array for the probabilities of the T thermodynamic states
    """
    new_therm_weights[:] = 0.0
    for i in range(len(bias_weight_sequences)):
        _mbar_direct_update_therm_weights(
            <int*> _np.PyArray_DATA(therm_state_counts),
            <double*> _np.PyArray_DATA(therm_weights),
            <double*> _np.PyArray_DATA(bias_weight_sequences[i]),
            therm_state_counts.shape[0],
            bias_weight_sequences[i].shape[0],
            <double*> _np.PyArray_DATA(new_therm_weights))
    new_therm_weights /= new_therm_weights[0]

def estimate_therm_energies(
    therm_state_counts, bias_energy_sequences,
    maxiter=1000, maxerr=1.0E-8, therm_energies=None,
    save_convergence_info=0, callback=None):
    r"""
    Estimate the thermodynamic free energies
        
    Parameters
    ----------
    therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        numbers of samples in the T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in free energies
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64), OPTIONAL
        initial guess for the reduced free energies of the T thermodynamic states
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    """
    T = therm_state_counts.shape[0]
    therm_state_counts = therm_state_counts.astype(_np.intc)
    log_therm_state_counts = _np.log(therm_state_counts)
    shift = _np.min([_np.min(b, axis=0) for b in bias_energy_sequences], axis=0)
    if therm_energies is None:
        therm_energies = _np.zeros(shape=(T,), dtype=_np.float64)
        therm_weights = _np.ones(shape=(T,), dtype=_np.float64)
    else:
        therm_weights = _np.exp(shift - therm_energies)
    bias_weight_sequences = [_np.exp(shift - b) for b in bias_energy_sequences]
    old_therm_energies = therm_energies.copy()
    old_therm_weights = therm_weights.copy()
    increments = []
    sci_count = 0
    for m in range(maxiter):
        sci_count += 1
        update_therm_weights(
            therm_state_counts, old_therm_weights, bias_weight_sequences, therm_weights)
        therm_energies = -_np.log(therm_weights)
        delta_therm_energies = _np.abs(therm_energies - old_therm_energies)
        err = _np.max(delta_therm_energies)
        if sci_count == save_convergence_info:
            sci_count = 0
            increments.append(err)
        if callback is not None:
            try:
                callback(iteration_step = m,
                         therm_weights = therm_weights,
                         old_therm_weights = old_therm_weights,
                         err=err,
                         maxerr=maxerr,
                         maxiter=maxiter)
            except CallbackInterrupt:
                break
        if err < maxerr:
            break
        else:
            old_therm_weights[:] = therm_weights[:]
            old_therm_energies[:] = therm_energies[:]
    therm_energies = shift - _np.log(therm_weights)
    if err >= maxerr:
        _warn("MBAR did not converge: last increment = %.5e" % err, _NotConvergedWarning)
    if save_convergence_info == 0:
        increments = None
    else:
        increments = _np.array(increments, dtype=_np.float64)
    return therm_energies, increments

def estimate(
    therm_state_counts, bias_energy_sequences, conf_state_sequences,
    maxiter=1000, maxerr=1.0E-8, therm_energies=None,
    n_conf_states=None, save_convergence_info=0, callback=None):
    r"""
    Estimate the (un)biased reduced free energies and thermodynamic free energies
        
    Parameters
    ----------
    therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        numbers of samples in the T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    conf_state_sequences : list of numpy.ndarray(shape=(X_i), dtype=numpy.intc)
        discrete state indices (cluster indices) for all X samples
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in free energies
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64), OPTIONAL
        initial guess for the reduced free energies of the T thermodynamic states
    n_conf_states : int, optional, default=None
        the number of configurational states in `conf_state_sequence`.
        If None, this is set to max(conf_state_sequence)+1.
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    conf_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced unbiased discrete state (cluster) free energies
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced discrete state free energies for all combinations of
        T thermodynamic states and M discrete states
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    """
    T = therm_state_counts.shape[0]
    therm_state_counts = therm_state_counts.astype(_np.intc)
    if n_conf_states is None:
        M = 1 + max([_np.max(c) for c in conf_state_sequences])
    else:
        M = n_conf_states
    assert len(conf_state_sequences) == len(bias_energy_sequences)
    for s, b in zip(conf_state_sequences, bias_energy_sequences):
        assert s.ndim == 1
        assert s.dtype == _np.intc
        assert b.ndim == 2
        assert b.dtype == _np.float64
        assert s.shape[0] == b.shape[0]
        assert b.shape[1] == T
        assert s.flags.c_contiguous
        assert b.flags.c_contiguous
    log_therm_state_counts = _np.log(therm_state_counts)
    shift = _np.min([_np.min(b, axis=0) for b in bias_energy_sequences], axis=0)
    if therm_energies is None:
        therm_energies = _np.zeros(shape=(T,), dtype=_np.float64)
        therm_weights = _np.ones(shape=(T,), dtype=_np.float64)
    else:
        therm_weights = _np.exp(shift - therm_energies)
    bias_weight_sequences = [_np.exp(shift - b) for b in bias_energy_sequences]
    scratch_M = _np.zeros(shape=(M,), dtype=_np.float64)
    scratch_T = _np.zeros(shape=(T,), dtype=_np.float64)
    therm_energies, increments = estimate_therm_energies(
        therm_state_counts, bias_energy_sequences,
        maxiter=maxiter, maxerr=maxerr, therm_energies=therm_energies,
        save_convergence_info=save_convergence_info, callback=callback)
    conf_energies, biased_conf_energies = _mbar.get_conf_energies(
        log_therm_state_counts, therm_energies,
        bias_energy_sequences, conf_state_sequences, scratch_T, M)
    _mbar.normalize(
        scratch_M, therm_energies, conf_energies, biased_conf_energies)
    return therm_energies, conf_energies, biased_conf_energies, increments



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import thermotools.util as util
import numpy as np
from nose.tools import assert_true
from numpy.testing import assert_array_equal, assert_almost_equal

####################################################################################################
#   sorting
####################################################################################################

def test_mixed_sort_reverse():
    # testing against numpy.sort()
    x = np.ascontiguousarray(np.arange(1000)[::-1].astype(np.float64))
    y = np.sort(x)
    util.mixed_sort(x)
    assert_array_equal(x, y)

def test_mixed_sort_random():
    # testing against numpy.sort()
    x = np.random.rand(1000).astype(np.float64)
    y = np.sort(x)
    util.mixed_sort(x)
    assert_array_equal(x, y)

####################################################################################################
#   direct summation schemes
####################################################################################################

def test_kahan_summation():
    # np.sum() fails for this array when unsorted
    array = np.array([1.0E-8, 1.0, 1.0E+8] * 100000, dtype=np.float64)
    result = util.kahan_summation(array, sort_array=False)
    assert_true(result == 10000000100000.001)
    result = util.kahan_summation(array, sort_array=True, inplace=False)
    assert_true(result == 10000000100000.001)
    result = util.kahan_summation(array, sort_array=True, inplace=True)
    assert_true(result == 10000000100000.001)

####################################################################################################
#   logspace summation schemes
####################################################################################################

def test_logsumexp_zeros():
    N = 10000
    data = np.zeros(shape=(N,), dtype=np.float64)
    assert_almost_equal(util.logsumexp(data, inplace=False), np.log(N), decimal=15)
    assert_almost_equal(util.logsumexp(-data, inplace=False), np.log(N), decimal=15)

def test_logsumexp_converged_geometric_series():
    data = np.ascontiguousarray(np.arange(10000)[::-1].astype(np.float64))
    assert_almost_equal(
        util.logsumexp(-data, inplace=False, sort_array=False, use_kahan=False),
        0.45867514538708193, decimal=15)
    assert_almost_equal(
        util.logsumexp(-data, inplace=False, sort_array=False, use_kahan=True),
        0.45867514538708193, decimal=15)
    assert_almost_equal(
        util.logsumexp(-data, inplace=False, sort_array=True, use_kahan=False),
        0.45867514538708193, decimal=15)
    assert_almost_equal(
        util.logsumexp(-data, inplace=False, sort_array=True, use_kahan=True),
        0.45867514538708193, decimal=15)
    assert_almost_equal(
        util.logsumexp(-data, inplace=True, sort_array=True, use_kahan=True),
        0.45867514538708193, decimal=15)
    
def test_logsumexp_truncated_diverging_geometric_series():
    data = np.ascontiguousarray(np.arange(10000)[::-1].astype(np.float64))
    assert_almost_equal(
        util.logsumexp(data, inplace=False, sort_array=False, use_kahan=False),
        9999.4586751453862, decimal=15)
    assert_almost_equal(
        util.logsumexp(data, inplace=False, sort_array=False, use_kahan=True),
        9999.4586751453862, decimal=15)
    assert_almost_equal(
        util.logsumexp(data, inplace=False, sort_array=True, use_kahan=False),
        9999.4586751453862, decimal=15)
    assert_almost_equal(
        util.logsumexp(data, inplace=False, sort_array=True, use_kahan=True),
        9999.4586751453862, decimal=15)
    assert_almost_equal(
        util.logsumexp(data, inplace=True, sort_array=True, use_kahan=True),
        9999.4586751453862, decimal=15)

def test_logsumexp_pair():
    assert_almost_equal(util.logsumexp_pair(0.0, 0.0), np.log(2.0), decimal=15)
    assert_almost_equal(util.logsumexp_pair(1.0, 1.0), 1.0 + np.log(2.0), decimal=15)
    assert_almost_equal(util.logsumexp_pair(10.0, 10.0), 10.0 + np.log(2.0), decimal=15)
    assert_almost_equal(util.logsumexp_pair(100.0, 100.0), 100.0 + np.log(2.0), decimal=15)
    assert_almost_equal(util.logsumexp_pair(1000.0, 1000.0), 1000.0 + np.log(2.0), decimal=15)
    assert_almost_equal(util.logsumexp_pair(10.0, 0.0), 10.000045398899218, decimal=15)
    assert_almost_equal(util.logsumexp_pair(0.0, 10.0), 10.000045398899218, decimal=15)
    assert_almost_equal(util.logsumexp_pair(100.0, 0.0), 100.0, decimal=15)
    assert_almost_equal(util.logsumexp_pair(0.0, 100.0), 100.0, decimal=15)
    assert_almost_equal(util.logsumexp_pair(1000.0, 0.0), 1000.0, decimal=15)
    assert_almost_equal(util.logsumexp_pair(0.0, 1000.0), 1000.0, decimal=15)

####################################################################################################
#   counting states and transitions
####################################################################################################

def test_break_points_us_like_trajs():
    X = 2000
    T = 100
    for K in range(T):
        bp = util.get_therm_state_break_points(np.ones(shape=(X,), dtype=np.intc) * K)
        assert_true(bp.shape[0] == 1)
        assert_true(bp[0] == 0)

def test_break_points_st_like_trajs():
    bp = util.get_therm_state_break_points(np.arange(1000).astype(np.intc))
    assert_true(bp.shape[0] == 1000)
    assert_array_equal(bp, np.array(range(1000), dtype=np.intc))
    bp = util.get_therm_state_break_points(
        np.array([0] * 10 + [1] * 20 + [0] * 30 + [1], dtype=np.intc))
    assert_true(bp.shape[0] == 4)
    assert_array_equal(bp, np.array([0, 10, 30, 60], dtype=np.intc))

def test_count_matrices_single_counts():
    dtrajs = [
        np.array([0, 0, 1, 1, 2, 2, 0, 2, 1, 0], dtype=np.intc),
        np.array([0, 0, 1, 1, 2, 2, 0, 2, 1, 0], dtype=np.intc)]
    ttrajs = [np.array([0] * 10, dtype=np.intc), np.array([1] * 10, dtype=np.intc)]
    # dtraj = [
    #     np.array(
    #         [[0, 0], [0, 0], [0, 1], [0, 1], [0, 2], [0, 2], [0, 0], [0, 2], [0, 1], [0, 0]],
    #         dtype=np.intc),
    #     np.array(
    #         [[1, 0], [1, 0], [1, 1], [1, 1], [1, 2], [1, 2], [1, 0], [1, 2], [1, 1], [1, 0]],
    #         dtype=np.intc)]
    C_K = util.count_matrices(ttrajs, dtrajs, 1, sparse_return=False)
    ref = np.ones(shape=(2, 3, 3), dtype=np.intc)
    assert_array_equal(C_K, ref)

def test_count_matrices_st_traj():
    ttraj = [np.array([0, 0, 1, 1, 1, 0, 0, 2, 2, 2, 0, 0], dtype=np.intc)]
    dtraj = [np.array([0, 0, 0, 1, 0, 1, 1, 1, 2, 1, 2, 2], dtype=np.intc)]
    C_K = util.count_matrices(ttraj, dtraj, 1, sliding=True, sparse_return=False, nthermo=4, nstates=4)
    ref = np.zeros(shape=(4, 4, 4), dtype=np.intc)
    ref[0, 0, 0] = 1
    ref[0, 1, 1] = 1
    ref[0, 2, 2] = 1
    ref[1, 0, 1] = 1
    ref[1, 1, 0] = 1
    ref[2, 1, 2] = 1
    ref[2, 2, 1] = 1
    assert_array_equal(C_K, ref)

def test_state_counts():
    ttrajs = [np.zeros(shape=(10,), dtype=np.intc), 2 * np.ones(shape=(20,), dtype=np.intc)]
    dtrajs = [np.zeros(shape=(10,), dtype=np.intc), 2 * np.ones(shape=(20,), dtype=np.intc)]
    ref = np.array([[10, 0, 0, 0], [0] * 4, [0, 0, 20, 0], [0] * 4, [0] * 4], dtype=np.intc)
    N = util.state_counts(ttrajs, dtrajs, nthermo=5, nstates=4)
    assert_array_equal(N, ref)

def test_restriction():
    T = 10
    M = 100
    X = 1000
    state_sequence = np.array([[0, i] for i in range(M)] * 10, dtype=np.intc)
    bias_energy_sequence = np.ascontiguousarray(
        np.array([[i] * T for i in range(X)], dtype=np.float64).transpose())
    cset = [i for i in range(M) if i % 2 == 0]
    ref_state_sequence = np.array([[0, i] for i in range(int(M / 2))] * 10, dtype=np.intc)
    ref_bias_energy_sequence = np.ascontiguousarray(
        np.array([[i] * T for i in range(X) if i % 2 == 0], dtype=np.float64).transpose())
    new_state_sequence, new_bias_energy_sequence = util.restrict_samples_to_cset(
        state_sequence, bias_energy_sequence, cset)
    assert_array_equal(new_state_sequence, ref_state_sequence)
    assert_array_equal(new_bias_energy_sequence, ref_bias_energy_sequence)

####################################################################################################
#   bias calculation tools
####################################################################################################

def test_get_umbrella_bias_binary():
    nsamples = 100
    nthermo = 2
    ndim = 3
    traj = np.linspace(0.0, 2.0, nsamples)
    for _i in range(1, ndim):
        traj = np.vstack((traj, np.linspace(0.0, 2.0, nsamples)))
    traj = np.ascontiguousarray(traj.T, dtype=np.float64)
    umbrella_centers = np.zeros(shape=(nthermo, ndim), dtype=np.float64)
    umbrella_centers[1, :] = 1.0
    force_constants = np.array([
        np.zeros(shape=(ndim, ndim), dtype=np.float64), np.eye(ndim, dtype=np.float64)])
    width = np.zeros(shape=(ndim,), dtype=np.float64)
    bias = util.get_umbrella_bias(traj, umbrella_centers, force_constants, width)
    ref = np.vstack((
        np.zeros(shape=(nsamples)),
        0.5 * ndim * np.linspace(-1.0, 1.0, nsamples)**2)).T.astype(np.float64)
    assert_almost_equal(bias, ref, decimal=15)

####################################################################################################
#   transition matrix renormalization
####################################################################################################



#!/usr/bin/env python

# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
thermotools is a lowlevel implementation toolbox for the analyis of multi-ensemble calculations.
"""

from setuptools import setup, Extension
import versioneer

def extensions():
    from numpy import get_include
    from Cython.Build import cythonize
    extra_compile_args = ["-O3", "-std=c99"]
    ext_bar = Extension(
        "thermotools.bar",
        sources=["ext/bar/bar.pyx", "ext/bar/_bar.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    ext_wham = Extension(
        "thermotools.wham",
        sources=["ext/wham/wham.pyx", "ext/wham/_wham.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    ext_mbar = Extension(
        "thermotools.mbar",
        sources=["ext/mbar/mbar.pyx", "ext/mbar/_mbar.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    ext_tram = Extension(
        "thermotools.tram",
        sources=["ext/tram/tram.pyx", "ext/tram/_tram.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    ext_dtram = Extension(
        "thermotools.dtram",
        sources=["ext/dtram/dtram.pyx", "ext/dtram/_dtram.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    ext_trammbar = Extension(
        "thermotools.trammbar",
        sources=["ext/trammbar/trammbar.pyx", "ext/tram/_tram.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args + ["-DTRAMMBAR"])
    ext_mbar_direct = Extension(
        "thermotools.mbar_direct",
        sources=["ext/mbar_direct/mbar_direct.pyx", "ext/mbar_direct/_mbar_direct.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    ext_tram_direct = Extension(
        "thermotools.tram_direct",
        sources=["ext/tram_direct/tram_direct.pyx", "ext/tram_direct/_tram_direct.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    ext_trammbar_direct = Extension(
        "thermotools.trammbar_direct",
        sources=["ext/trammbar_direct/trammbar_direct.pyx", "ext/tram_direct/_tram_direct.c", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args + ["-DTRAMMBAR"])
    ext_util = Extension(
        "thermotools.util",
        sources=["ext/util/util.pyx", "ext/util/_util.c"],
        include_dirs=[get_include()],
        extra_compile_args=extra_compile_args)
    exts = [
        ext_bar,
        ext_wham,
        ext_mbar,
        ext_tram,
        ext_dtram,
        ext_trammbar,
        ext_mbar_direct,
        ext_tram_direct,
        ext_trammbar_direct,
        ext_util]
    return cythonize(exts)

class lazy_cythonize(list):
    """evaluates extension list lazyly.
    pattern taken from http://tinyurl.com/qb8478q"""
    def __init__(self, callback):
        self._list, self.callback = None, callback
    def c_list(self):
        if self._list is None: self._list = self.callback()
        return self._list
    def __iter__(self):
        for e in self.c_list(): yield e
    def __getitem__(self, ii): return self.c_list()[ii]
    def __len__(self): return len(self.c_list())

long_description = 'The thermotools package is a lowlevel implementation toolbox for the analyis \
of multi-ensemble calculations. It contains estimators for the state-continuous transition-based \
reweighting analysis method (TRAM) and its state-discrete variant (dTRAM), Bennet acceptance ratio \
(BAR) and its multi-state variant (MBAR), and the weighted histogram analysis method (WHAM). \
While you can use thermotools on its own, we recommend to use it in combination with PyEMMA.'

setup(
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=lazy_cythonize(extensions),
    name='thermotools',
    version=versioneer.get_version(),
    description='lowlevel implementation toolbox for the analyis of multi-ensemble calculations',
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics'],
    keywords=[
        'MEMM',
        'multi ensemble',
        'free energy',
        'Markov state model',
        'BAR',
        'WHAM',
        'MBAR',
        'TRAM',
        'dTRAM'],
    url='https://github.com/markovmodel/thermotools',
    maintainer='Christoph Wehmeyer',
    maintainer_email='christoph.wehmeyer@fu-berlin.de',
    license='LGPLv3+',
    setup_requires=[
        'numpy>=1.7',
        'cython>=0.20',
        'setuptools>=0.6'],
    tests_require=[
        'numpy>=1.7',
        'scipy>=0.11',
        'msmtools>=1.1.3',
        'nose'],
    install_requires=[
        'numpy>=1.7',
        'scipy>=0.11',
        'msmtools>=1.1.3'],
    packages=['thermotools'],
    test_suite='nose.collector',
    scripts=[]
)



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the TRAM estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np
import sys
from thermotools import trammbar as _trammbar

from warnings import warn as _warn
from msmtools.util.exceptions import NotConvergedWarning as _NotConvergedWarning
from thermotools.trammbar import get_pointwise_unbiased_free_energies, estimate_transition_matrix, estimate_transition_matrices

from .callback import CallbackInterrupt

__all__ = [
    'estimate_transition_matrix',
    'estimate_transition_matrices',
    'estimate',
    'get_pointwise_unbiased_free_energies']

DEF TRAMMBAR = True

cdef extern from "../tram_direct/_tram_direct.h":
    void _tram_direct_update_lagrangian_mult(
        double *lagrangian_mult, double *biased_conf_weights,
        int *count_matrices, int* state_counts, int n_therm_states, int n_conf_states,
        double *new_lagrangian_mult)
    void _tram_direct_get_Ref_K_i(
        double *lagrangian_mult, double *biased_conf_weights, int *count_matrices,
        int *state_counts, int n_therm_states, int n_conf_states, double *R_K_i,
        # TRAMMBAR below
        double *therm_energies, int *equilibrium_therm_state_counts,
        double overcounting_factor)
    void _tram_direct_update_biased_conf_weights(
        double *bias_sequence, int *state_sequence, int seq_length, double *R_K_i,
        int n_therm_states, int n_conf_states, double *new_biased_conf_weights)
    void _tram_direct_dtram_like_update(
        double *lagrangian_mult, double *biased_conf_weights,
        int *count_matrices, int *state_counts, int n_therm_states, int n_conf_states,
        double *scratch_M, int *scratch_M_int, double *new_biased_conf_weights)

def update_lagrangian_mult(
    _np.ndarray[double, ndim=2, mode="c"] lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_weights not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=2, mode="c"] new_lagrangian_mult not None):
    _tram_direct_update_lagrangian_mult(
        <double*> _np.PyArray_DATA(lagrangian_mult),
        <double*> _np.PyArray_DATA(biased_conf_weights),
        <int*> _np.PyArray_DATA(count_matrices),
        <int*> _np.PyArray_DATA(state_counts),
        lagrangian_mult.shape[0],
        lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(new_lagrangian_mult))

def get_Ref_K_i(
    _np.ndarray[double, ndim=2, mode="c"] lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_weights not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=2, mode="c"] R_K_i not None,
    # TRAMMBAR below
    _np.ndarray[double, ndim=1, mode="c"] therm_weights=None,
    _np.ndarray[int, ndim=1, mode="c"] equilibrium_therm_state_counts=None,
    double overcounting_factor=1.0):
    r"""
    Computes the product of TRAM pseudo-counts and biased_conf_weights.

    Parameters
    ----------
    lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        Lagrangian multipliers
    biased_conf_weights : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        exp(-reduced free energies)
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        number of visits to thermodynamic state K and Markov state i
    R_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        target array for product of TRAM pseudo-counts and biased_conf_weights
    """
    _tram_direct_get_Ref_K_i(
        <double*> _np.PyArray_DATA(lagrangian_mult),
        <double*> _np.PyArray_DATA(biased_conf_weights),
        <int*> _np.PyArray_DATA(count_matrices),
        <int*> _np.PyArray_DATA(state_counts),
        lagrangian_mult.shape[0],
        lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(R_K_i),
        <double*> _np.PyArray_DATA(therm_weights) if therm_weights is not None else NULL,
        <int*> _np.PyArray_DATA(equilibrium_therm_state_counts) if equilibrium_therm_state_counts is not None else NULL,
        overcounting_factor)

def update_biased_conf_weights(
    _np.ndarray[double, ndim=2, mode="c"] lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_weights not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    bias_weight_sequences,
    state_sequences,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=2, mode="c"] R_K_i not None,
    _np.ndarray[double, ndim=2, mode="c"] new_biased_conf_weights not None,
    # TRAMMBAR below
    _np.ndarray[double, ndim=1, mode="c"] therm_weights=None,
    equilibrium_bias_weight_sequences=None,
    equilibrium_state_sequences=None,
    _np.ndarray[int, ndim=1, mode="c"] equilibrium_therm_state_counts=None,
    double overcounting_factor=1.0):

    new_biased_conf_weights[:] = 0.0
    get_Ref_K_i(lagrangian_mult, biased_conf_weights, count_matrices,
                state_counts, R_K_i,
                therm_weights=therm_weights,
                equilibrium_therm_state_counts=equilibrium_therm_state_counts,
                overcounting_factor=overcounting_factor)
    for i in range(len(bias_weight_sequences)):
        _tram_direct_update_biased_conf_weights(
            <double*> _np.PyArray_DATA(bias_weight_sequences[i]),
            <int*> _np.PyArray_DATA(state_sequences[i]),
            state_sequences[i].shape[0],
            <double*> _np.PyArray_DATA(R_K_i),
            lagrangian_mult.shape[0],
            lagrangian_mult.shape[1],
            <double*> _np.PyArray_DATA(new_biased_conf_weights))
    if TRAMMBAR:
        if equilibrium_bias_weight_sequences is not None:
            new_biased_conf_weights *= overcounting_factor
            for i in range(len(equilibrium_bias_weight_sequences)):
                _tram_direct_update_biased_conf_weights(
                    <double*> _np.PyArray_DATA(equilibrium_bias_weight_sequences[i]),
                    <int*> _np.PyArray_DATA(equilibrium_state_sequences[i]),
                    equilibrium_state_sequences[i].shape[0],
                    <double*> _np.PyArray_DATA(R_K_i),
                    lagrangian_mult.shape[0],
                    lagrangian_mult.shape[1],
                    <double*> _np.PyArray_DATA(new_biased_conf_weights))

def dtram_like_update(
    _np.ndarray[double, ndim=2, mode="c"] lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_weights not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[int, ndim=1, mode="c"] scratch_M_int not None,
    _np.ndarray[double, ndim=2, mode="c"] new_biased_conf_weights not None):
    _tram_direct_dtram_like_update(
        <double*> _np.PyArray_DATA(lagrangian_mult),
        <double*> _np.PyArray_DATA(biased_conf_weights),
        <int*> _np.PyArray_DATA(count_matrices),
        <int*> _np.PyArray_DATA(state_counts),
        lagrangian_mult.shape[0],
        lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <int*> _np.PyArray_DATA(scratch_M_int),
        <double*> _np.PyArray_DATA(new_biased_conf_weights))

def estimate(
    count_matrices, state_counts, bias_energy_sequences, state_sequences,
    maxiter=1000, maxerr=1.0E-8, save_convergence_info=0,
    biased_conf_energies=None, log_lagrangian_mult=None, callback=None,
    N_dtram_accelerations=0,
    equilibrium_therm_state_counts=None,
    equilibrium_bias_energy_sequences=None, equilibrium_state_sequences=None,
    overcounting_factor = 1.0):
    r"""
    Estimate the reduced discrete state free energies and thermodynamic free energies

    Parameters
    ----------
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        transition count matrices for all T thermodynamic states
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        state counts for all M discrete and T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    state_sequences : list of numpy.ndarray(shape=(X_i), dtype=numpy.float64)
        discrete state indices for all X samples
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in free energies
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment
        and the actual loglikelihood
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64), OPTIONAL
        initial guess for the reduced discrete state free energies for all T thermodynamic states
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64), OPTIONAL
        initial guess for the logarithm of the Lagrangian multipliers

    Returns
    -------
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced discrete state free energies for all T thermodynamic states
    conf_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced unbiased discrete state free energies
    therm_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced thermodynamic free energies
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        logarithm of the Lagrangian multipliers
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    loglikelihoods : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of loglikelihoods     
    """
    if not TRAMMBAR:
        assert equilibrium_therm_state_counts is None
    
    n_therm_states = count_matrices.shape[0]
    n_conf_states = count_matrices.shape[1]
    assert len(state_sequences)==len(bias_energy_sequences)
    for s, b in zip(state_sequences, bias_energy_sequences):
        assert s.ndim == 1
        assert s.dtype == _np.intc
        assert b.ndim == 2
        assert b.dtype == _np.float64
        assert s.shape[0] == b.shape[0]
        assert b.shape[1] == n_therm_states
        assert s.flags.c_contiguous
        assert b.flags.c_contiguous
    if TRAMMBAR:
        assert N_dtram_accelerations == 0
        if equilibrium_state_sequences is not None:
            assert len(equilibrium_state_sequences) == len(equilibrium_bias_energy_sequences)
            for s, b in zip(equilibrium_state_sequences, equilibrium_bias_energy_sequences):
                assert s.ndim == 1
                assert s.dtype == _np.intc
                assert b.ndim == 2
                assert b.dtype == _np.float64
                assert s.shape[0] == b.shape[0]
                assert b.shape[1] == count_matrices.shape[0]
                assert s.flags.c_contiguous
                assert b.flags.c_contiguous
    else:
        assert equilibrium_bias_energy_sequences is None
        assert equilibrium_state_sequences is None

    assert(_np.all(
        state_counts >= _np.maximum(count_matrices.sum(axis=1), count_matrices.sum(axis=2))))

    # init lagrangian multipliers
    if log_lagrangian_mult is None:
        lagrangian_mult = 0.5 * (count_matrices + _np.transpose(
            count_matrices, axes=(0, 2, 1))).sum(axis=2).astype(_np.float64)
    else:
        lagrangian_mult = _np.exp(log_lagrangian_mult)
    # exploit invariance w.r.t. simultaneous scaling of energies and free energies
    # standard quantities-> scaled quantities
    # bias_energy^k(x)   -> bias_energy^k(x) - alpha^k
    # free_energy_i^k    -> free_energy_i^k - alpha^k
    # log \tilde{R}_i^k  -> log \tilde{R}_i^k - alpha^k
    if TRAMMBAR:
        if equilibrium_bias_energy_sequences is not None:
            all_bias_energy_sequences = bias_energy_sequences + equilibrium_bias_energy_sequences
        else:
            all_bias_energy_sequences = bias_energy_sequences
    else:
        all_bias_energy_sequences = bias_energy_sequences
    shift = _np.min([_np.min(b, axis=0) for b in all_bias_energy_sequences], axis=0) # minimum energy for every th. state        
    # init weights
    if biased_conf_energies is not None:
        biased_conf_weights = _np.exp(shift[:, _np.newaxis] - biased_conf_energies)
    else:
        biased_conf_weights = _np.ones(shape=(n_therm_states, n_conf_states), dtype=_np.float64)
        biased_conf_energies = _np.zeros(shape=(n_therm_states, n_conf_states), dtype=_np.float64)
    # init Boltzmann factors # TODO: offer in-place option
    bias_weight_sequences = [_np.exp(shift - b) for b in bias_energy_sequences]
    if TRAMMBAR:
        if equilibrium_bias_energy_sequences is not None:
            equilibrium_bias_weight_sequences = [_np.exp(shift - b) for b in equilibrium_bias_energy_sequences]
        else:
            equilibrium_bias_weight_sequences = None
    else:
        equilibrium_bias_weight_sequences =None
    increments = []
    loglikelihoods = []
    sci_count = 0
    R_K_i = _np.zeros(shape=(n_therm_states, n_conf_states), dtype=_np.float64)
    scratch_TM = _np.zeros(shape=(n_therm_states, n_conf_states), dtype=_np.float64)
    scratch_TM2 = _np.zeros(shape=(n_therm_states, n_conf_states), dtype=_np.float64)
    scratch_MM = _np.zeros(shape=(n_conf_states, n_conf_states), dtype=_np.float64)
    scratch_T = _np.zeros(shape=n_therm_states, dtype=_np.float64)
    scratch_M = _np.zeros(shape=n_conf_states, dtype=_np.float64)
    scratch_M_int = _np.zeros(shape=n_conf_states, dtype=_np.intc)
    occupied = _np.where(state_counts > 0)
    if _np.any(_np.isinf(biased_conf_energies[occupied])):
        print >>sys.stderr, 'Warning: detected inf in biased_conf_energies.' # TODO: possible Python3 violation
    partition_funcs = biased_conf_weights.sum(axis=1)
    old_partition_funcs = partition_funcs.copy()
    old_biased_conf_weights = biased_conf_weights.copy()
    old_lagrangian_mult = lagrangian_mult.copy()
    old_biased_conf_energies = biased_conf_energies.copy()
    old_stat_vectors = _np.zeros(shape=state_counts.shape, dtype=_np.float64)
    old_therm_energies = _np.zeros(shape=count_matrices.shape[0], dtype=_np.float64)
    for m in range(maxiter):
        sci_count += 1
        update_lagrangian_mult(
            old_lagrangian_mult, biased_conf_weights, count_matrices, state_counts, lagrangian_mult)
        update_biased_conf_weights(
            lagrangian_mult, old_biased_conf_weights, count_matrices, bias_weight_sequences,
            state_sequences, state_counts, R_K_i, biased_conf_weights,
            therm_weights=old_partition_funcs,
            equilibrium_bias_weight_sequences=equilibrium_bias_weight_sequences,
            equilibrium_state_sequences=equilibrium_state_sequences,
            equilibrium_therm_state_counts=equilibrium_therm_state_counts,
            overcounting_factor=overcounting_factor)
        for _n  in range(N_dtram_accelerations):
            old_biased_conf_weights[:] = biased_conf_weights[:]
            dtram_like_update(
                lagrangian_mult, old_biased_conf_weights, count_matrices,
                state_counts, scratch_M, scratch_M_int, biased_conf_weights)
        partition_funcs[:] = biased_conf_weights.sum(axis=1)
        stat_vectors = biased_conf_weights / partition_funcs[:, _np.newaxis]
        therm_energies = -_np.log(partition_funcs)
        delta_therm_energies = _np.abs(therm_energies - old_therm_energies)
        delta_stat_vectors =  _np.abs(stat_vectors - old_stat_vectors)
        err = max(_np.max(delta_therm_energies), _np.max(delta_stat_vectors[occupied]))
        if sci_count == save_convergence_info:
            sci_count = 0
            increments.append(err)
            with _np.errstate(divide='ignore'):
                log_lagrangian_mult = _np.log(lagrangian_mult)
                biased_conf_energies = shift[:, _np.newaxis] - _np.log(biased_conf_weights) # can contain -inf for empty state
            logL = _trammbar.log_likelihood_lower_bound(
                log_lagrangian_mult, biased_conf_energies, count_matrices,
                bias_energy_sequences, state_sequences, state_counts,
                scratch_TM2, scratch_M, scratch_T, scratch_TM, scratch_MM,
                therm_energies=therm_energies,
                equilibrium_bias_energy_sequences=equilibrium_bias_energy_sequences,
                equilibrium_state_sequences=equilibrium_state_sequences,
                equilibrium_therm_state_counts=equilibrium_therm_state_counts,
                overcounting_factor=overcounting_factor)
            loglikelihoods.append(logL)
        if callback is not None:
            try:
                callback(iteration_step = m,
                         biased_conf_weights = biased_conf_weights,
                         lagrangian_mult = lagrangian_mult,
                         old_biased_conf_weights = old_biased_conf_weights,
                         old_lagrangian_mult = old_lagrangian_mult,
                         occupied = occupied,
                         shift = shift,
                         err = err,
                         maxerr = maxerr,
                         maxiter = maxiter)
            except CallbackInterrupt:
                break
        if err < maxerr:
            break
        else:
            normalization_factor = _np.max(biased_conf_weights)
            biased_conf_weights /= normalization_factor
            old_lagrangian_mult[:] = lagrangian_mult[:]
            old_biased_conf_weights[:] = biased_conf_weights[:]
            old_therm_energies[:] = therm_energies[:] + _np.log(normalization_factor)
            old_partition_funcs[:] = partition_funcs / normalization_factor
            old_stat_vectors[:] = stat_vectors[:]
    with _np.errstate(divide='ignore'):
        biased_conf_energies = shift[:, _np.newaxis] - _np.log(biased_conf_weights) # can contain -inf for empty states
        log_lagrangian_mult = _np.log(lagrangian_mult)
        log_R_K_i = _np.log(R_K_i) + shift[:, _np.newaxis]
    conf_energies = _trammbar.get_conf_energies(
        bias_energy_sequences, state_sequences, log_R_K_i, scratch_T,
        equilibrium_bias_energy_sequences=equilibrium_bias_energy_sequences,
        equilibrium_state_sequences=equilibrium_state_sequences,
        overcounting_factor=overcounting_factor)
    therm_energies = _trammbar.get_therm_energies(biased_conf_energies, scratch_M)
    _trammbar.normalize(conf_energies, biased_conf_energies, therm_energies, scratch_M)
    if err >= maxerr:
        _warn("TRAM did not converge: last increment = %.5e" % err, _NotConvergedWarning)
    if save_convergence_info == 0:
        increments = None
        loglikelihoods = None
    else:
        increments = _np.array(increments, dtype=_np.float64)
        loglikelihoods = _np.array(loglikelihoods, dtype=_np.float64)
    return biased_conf_energies, conf_energies, therm_energies, \
        log_lagrangian_mult, increments, loglikelihoods
